package be.iminds.ilabt.jfed.ui.cli2.logging;

import be.iminds.ilabt.jfed.util.common.IOUtils;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.SerializationFeature;
import io.dropwizard.jackson.Jackson;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.io.PrintStream;
import java.util.List;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.function.Consumer;

public class ActionLogger {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(ActionLogger.class);
    private static final ObjectMapper MAPPER = Jackson.newObjectMapper();

    public interface ActionLogListener {
        void onOut(@Nonnull String s);
        void onJsonOut(@Nonnull Object o);
        void onErr(@Nonnull String s);
    }
    public static class StreamActionLogListener implements ActionLogListener{
        @Nonnull private final PrintStream outStream;
        @Nonnull private final PrintStream errStream;
        private final boolean doNonJsonOut;
        private final boolean doJsonOut;

        public StreamActionLogListener(@Nonnull PrintStream outStream, @Nonnull PrintStream errStream,
                                       boolean doNonJsonOut, boolean doJsonOut) {
            this.outStream = outStream;
            this.errStream = errStream;
            this.doNonJsonOut = doNonJsonOut;
            this.doJsonOut = doJsonOut;
        }

        public void onJsonOut(@Nonnull Object o) {
            if (!doJsonOut) return;
            String s;
            try {
                s = MAPPER
                        .writerWithDefaultPrettyPrinter()
                        .withFeatures(SerializationFeature.ORDER_MAP_ENTRIES_BY_KEYS)
                        .writeValueAsString(o);
            } catch (JsonProcessingException e) {
                LOG.error("Error converting action output to JSON", e);
                s = "Error converting action output to JSON: "+e.getMessage();
            }
            outStream.print(s);
            outStream.flush();
        }
        public void onOut(@Nonnull String s) {
            if (!doNonJsonOut) return;
            outStream.print(s);
            outStream.flush();
        }
        public void onErr(@Nonnull String s) {
            errStream.print(s);
            errStream.flush();
        }
    }

    private final List<ActionLogListener> listeners = new CopyOnWriteArrayList<>();

    private final AtomicInteger exceptionCount = new AtomicInteger(0);

    @Nullable private final Thread.UncaughtExceptionHandler uncaughtExceptionHandler;

    public ActionLogger() {
        this.uncaughtExceptionHandler = Thread.currentThread().getUncaughtExceptionHandler();
    }
    public ActionLogger(@Nullable Thread.UncaughtExceptionHandler uncaughtExceptionHandler) {
        this.uncaughtExceptionHandler = uncaughtExceptionHandler;
    }


    public void actionJsonOut(@Nonnull Object o) {
        executeSafeForAllListeners((l) -> l.onJsonOut(o));
    }
    public void actionPrint(@Nonnull String s) {
        executeSafeForAllListeners((l) -> l.onOut(s));
    }
    public void actionPrintln(@Nonnull String s) {
        executeSafeForAllListeners((l) -> l.onOut(s+"\n"));
    }

    public void errorPrintln(@Nonnull String s) {
        executeSafeForAllListeners((l) -> l.onErr(s+"\n"));
    }
    public void errorPrintln(String msg, Throwable throwable) {
        String stackTrace = throwable == null ? "" : "\n\n"+IOUtils.exceptionToStacktraceString(throwable);
        errorPrintln(msg+stackTrace);
    }

    public void addESpecLogListener(@Nonnull ActionLogListener listener) {
        listeners.add(listener);
    }
    public void removeESpecLogListener(@Nonnull ActionLogListener listener) {
        listeners.remove(listener);
    }



    private void executeSafeForAllListeners(Consumer<ActionLogListener> c) {
        for (ActionLogListener l : listeners) {
            try {
                c.accept(l);
            } catch (Exception e) {
                handleExceptionInListener(e);
            }
        }
    }

    /**
     * We only handle the first exception with an uncaughtExceptionhandler, if present.
     * Otherwise, the user might see too much similar exceptions.
     *
     * @param e
     */
    private void handleExceptionInListener(Exception e) {
        int count = exceptionCount.getAndIncrement();
        if (count == 0 && uncaughtExceptionHandler != null) {
            try {
                uncaughtExceptionHandler.uncaughtException(Thread.currentThread(), e);
            } catch (Exception e2) {
                LOG.error("Ignoring exception in uncaughtExceptionHandler", e2);
            }
        } else {
            LOG.error("Ignoring exception "+count+" in ESpecLogger listener", e);
        }
    }

}
